/*
 * Copyright (c) 2003-2007, Takahide Matsutsuka.
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by Takahide Matsutsuka."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 * 4. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.  
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Intel HEX format to PC-6001 CAS format conversion utility.
 */

#include <stdio.h>
#include "hex2cas.h"

// output file structure
// 1/ prefix (prefix_length bytes)
// 2/ startup (STARTUP_LENGTH bytes)
// 3/ a little patch for rst8 (3 bytes)
// 4/ codeseg

static unsigned char prefix_p6[] = {
	0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 
	0xd3, 0xd3, 0x47, 0x6f, 0x00, 0x00, 0x00, 0x00, 
	0x0d, 0x84, 0x0a, 0x00, 0xa5, '&',  'H',  '8',
	'4',  '0',  'F',  0x00, 0x00, 0x00
};
static int prefix_length_p6 = 0x1e;

static unsigned char prefix_p88[] = {
	0xd3,0xd3,0xd3,0xd3,0xd3,0xd3,0xd3,0xd3,0xd3,0xd3,0x64,0x65,0x66,0x61,0x6c,0x74,
	0x0d,0x00,0x0a,0x00,0x97,0x20,0xe0,0xf1,0x0c,0x00,0x01,0x00,0x18,0x00,0x14,0x00,
	0x41,0xf1,0xe0,0x28,0x11,0x29,0x00,0x00,0x00,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
	0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20
};
static int prefix_length_p88 = 0x10e;

//  char prefix_p88[] = { 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 
//  		      0xd3, 0xd3, 0x47, 0x6f, 0x00, 0x00, 0x00, 0x00, 
//        /* 0x0001 */    0x0d, 0x00, 0x0a, 0x00, 0x97, 0x20, 0xe0, 0xf1,
//  		      0x0c, 0x1a, 0x00, 0x00, 0x18, 0x00, 0x14, 0x00,
//  		      //    ~~~~~~~~~~ start address
//        /* 0x0011 */    0x41, 0xf1, 0xe0, 0x28, 0x11, 0x29, 0x00, 0x00,
//                        0x00 };  /* followed by 12 bytes long startup code */
//  //                       8     9    10    11    12  (here)

unsigned char prefix_rom[] = { 'A', 'B', 0x04, 0x40 };
int prefix_length_rom = 0x4;

static int suffix_length = 12;
static char suffix[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 
			 0x00, 0x00, 0x00, 0x00 };

/**
 * 64kB size memory map.
 */
static unsigned char memory[65536];

/**
 * Maximum address position to be used.
 */
static int maxAddress;

/**
 * Convert a character to a value.
 * @param ch a character to convert
 * @return integer value represents the given character
 */
static int aton(const unsigned char ch) {
  int n;
  if (ch >= '0' && ch <= '9') {
    return ch - '0';
  }
  if (ch >= 'A' && ch <= 'F') {
    return ch - 'A' + 10;
  }
  if (ch >= 'a' && ch <= 'f') {
    return ch - 'a' + 10;
  }

  return 0;
}

/**
 * Convert a byte characters from a given file and returns a byte.
 * @param in file
 * @return -1 if EOF
 */
static int getByte(FILE *in) {
  int ch1, ch2;
  ch1 = getc(in);
  if (ch1 == EOF) {
    return -1;
  }
  ch2 = getc(in);
  if (ch2 == EOF) {
    return -1;
  }
  return 16 * aton(ch1) + aton(ch2);
}

/**
 * Extract a 64kB memory map from given file.
 * IHEX format is as follows:
 * :A_B___C_D_....D_E_
 * A_  : size of this chunk
 * B___: address (big endian)
 * C_  : record type
 * D_....D_: data
 * E_  : check sum
 * @param inFilename file name to convert
 * @return 0 if noerror, otherwise if error
 */
int hex2mem(const char *inFilename) {
  FILE *in;
  maxAddress = 0;

  memset(memory, 0, 65536);

  in = fopen(inFilename, "rb");
  if (in == NULL) {
    printf("cannot open input file\n");
    return 1;
  }

  while(1) {
    int ch;
    int length;
    int address;
    int tmp;

    while ((ch = getc(in)) != ':') {
      if (ch == EOF) {
        break;
      }
    }
    if (ch == EOF) {
      break;
    }

    // get length of this chunk
    length = getByte(in);
    if (length <= 0) {
      break;
    }

    // make an address
    tmp = getByte(in);
    if (tmp < 0) {
      break;
    }
    address = tmp * 256;
    tmp = getByte(in);
    if (tmp < 0) {
      break;
    }
    address += tmp;    
    if (maxAddress < (address + length)) {
      maxAddress = address + length;
    }

    // ignore record type
    if (getByte(in) < 0) {
      break;
    }

    while (length > 0) {
      memory[address] = getByte(in);
      address++;
      length--;
    }
  }

  fclose(in);
  return 0;
}

/**
 * Emit the memory map to an output file. 
 * @param outFilename file name to emit
 * @param archFilename file name to embed the output file
 * @param mode translation mode
 * @return 0 if noerror, otherwise if error
 */
int mem2cas(const char *outFilename, const char archFilename[6],
	    const unsigned char mode) {
  FILE *out;
  int ch;
  int i;
  int codeseg;
  unsigned char* prefix;
  int prefix_length;
  
  if (archFilename[0] == 0) {
    archFilename = DEFAULT_ARCH_FILENAME;
  }

  // MSB of BASIC RAM start address
  switch (mode) {
  case MODE1:
    prefix = prefix_p6;
    prefix_length = prefix_length_p6;
    prefix[17] = 0xc4;
    prefix[23] = 'C';
    codeseg = MODE1_ADDRESS_CODESEG;
    break;
  case MODE5:
    prefix = prefix_p6;
    prefix_length = prefix_length_p6;
    prefix[17] = 0x80;
    prefix[24] = '0';
    codeseg = MODE5_ADDRESS_CODESEG;
    break;
  case MODEN88:
    prefix = prefix_p88;
    prefix_length = prefix_length_p88;
    codeseg = MODEN88_ADDRESS_CODESEG;
    break;
  case MODEROM:
    prefix = prefix_rom;
    prefix_length = prefix_length_rom;
    codeseg = MODEROM_ADDRESS_CODESEG;
    break;
  case MODE2:
  default:
    prefix = prefix_p6;
    prefix_length = prefix_length_p6;
    prefix[23] = 0x38;
    prefix[24] = 0x34;
    codeseg = MODE2_ADDRESS_CODESEG;
    break;
  }
  // set p6 file name
  for (i = 0; i < 6; i++) {
    prefix[i + 0x0a] = archFilename[i];
  }

  // open a file
  out = fopen(outFilename, "wb");
  if (out == NULL) {
    printf("cannot open output file\n");
    return 1;
  }

  // write prefix
  for (i = 0; i < prefix_length; i++) {
    putc(prefix[i], out);
  }

  // write codeseg
  for (i = codeseg; i < maxAddress; i++) {
    putc(memory[i], out);
  }
  if (mode != MODEROM) {
    // write suffix
    for (i = 0; i < suffix_length; i++) {
      putc(suffix[i], out);
    }
  } else {
    // write trailing bytes
    for (; i < 0x8000; i++) {
      putc(memory[i], out);
    }
  }

  fclose(out);

  return 0;
}
